"""Base hook definitions"""

import functools
import logging

from .. import ConfigurationError
from ..components import get_entrypoint

_logger = logging.getLogger(__name__)


class Hook:
    """Helper class to manage hooks

    Parameters
    ----------
    name: str
        Hook name
    """

    #: Entrypoint namespace
    _namespace = "troika.hooks"

    registered_hooks = {}

    @classmethod
    def declare(cls, func, name=None):
        """Register a hook

        Parameters
        ----------
        func: callable
            Hook specification function (code is discarded)
        name: str or None
            Name of the hook (default: ``func.__name__``)
        """
        if name is None:
            name = func.__name__
        hook = cls(name)
        functools.update_wrapper(hook, func)
        cls.registered_hooks[name] = hook
        return hook

    def __init__(self, name):
        self.name = name
        self._hooks = {}
        self._impl = None

    def __call__(self, *args, **kwargs):
        if self._impl is None:
            raise ValueError("Attempt to call a non-instantiated hook registry")
        _logger.debug("Executing %s hooks", self.name)
        results = []
        for funcname, func in self._impl:
            _logger.debug("Calling hook function %s", funcname)
            res = func(*args, **kwargs)
            results.append(res)
        return results

    def instantiate(self, hooks):
        """Select the requested hook implementations

        Parameters
        ----------
        hooks: list of str
            Requested hooks
        """
        hookfuncs = []
        for hookname in hooks:
            try:
                hookfunc = get_entrypoint(f"{self._namespace}.{self.name}", hookname)
            except ValueError:
                msg = f"Implementation {hookname!r} not found for {self.name} hook"
                raise ConfigurationError(msg)
            hookfuncs.append((hookname, hookfunc))
        self._impl = hookfuncs


@Hook.declare
def at_startup(action, site, args):
    """Startup hook

    Parameters
    ----------
    action: {"submit", "monitor", "kill"}
        Action that was requested
    site: :py:class:`troika.sites.base.Site`
        Selected site
    args: :py:class:`argparse.Namespace`-like
        Command-line arguments

    Returns
    -------
    bool
        If True, interrupt the action
    """


@Hook.declare
def pre_submit(site, script, output, dryrun):
    """Pre-submit hook

    Parameters
    ----------
    site: :py:class:`troika.sites.base.Site`
        Selected site
    script: path-like
        Path to the script to be submitted
    output: path-like
        Path to the job output file
    dryrun: bool
        If True, do not do anything, only print actions
    """


@Hook.declare
def post_kill(site, script, output, jid, cancel_status, dryrun):
    """Post-kill hook

    Parameters
    ----------
    site: :py:class:`troika.sites.base.Site`
        Selected site
    script: path-like
        Path to the script file of the job being killed
    output: path-like or None
        Path to the job output file
    jid: str
        Job ID of the job being killed
    cancel_status: str
        CANCELLED, KILLED or TERMINATED
    dryrun: bool
        If True, do not do anything, only print actions
    """


@Hook.declare
def at_exit(action, site, args, sts, logfile):
    """Exit hook

    Parameters
    ----------
    action: {"submit", "monitor", "kill"}
        Action that was requested
    site: :py:class:`troika.sites.base.Site`
        Selected site
    args: :py:class:`argparse.Namespace`-like
        Command-line arguments
    sts: int
        Exit status
    logfile: path-like
        Path to the log file
    """



"""SGE-managed site"""

import locale
import logging
import pathlib
import re
from collections import OrderedDict

from .. import InvocationError, RunError
from ..connection import PIPE
from ..parser import BaseParser
from ..utils import check_retcode, command_as_list
from .base import Site

_logger = logging.getLogger(__name__)


def _split_sge_directive(arg):
    """Split the argument of a SGE directive

    >>> _split_sge_directive(b"-o foo")
    (b'-o', b'foo')
    >>> _split_sge_directive(b"-N job")
    (b'-N', b'job')
    >>> _split_sge_directive(b"-V")
    (b'-V', None)
    """
    m = re.match(rb"(\S+)(\s+)?(.*)?$", arg)
    if m is None:
        raise RunError(r"Malformed qsub argument: {arg!r}")
    key, sep, val = m.groups()
    if sep is None:
        assert val == b""
        val = None
    return key, val


class SGEDirectiveParser(BaseParser):
    """Parser that processes a script to extract SGE directives

    Parameters
    ----------
    drop_keys: Iterable[bytes]
        Directives to ignore, e.g. ``[b'-o', b'-e']``

    Members
    -------
    data: collections.OrderedDict[bytes, (bytes or None, bytes)]
        Directives that have been parsed. The first item of the dict value is
        the parsed value, if any, and the second is the full line, including
        the line terminator.
    """

    DIRECTIVE_RE = re.compile(rb"^#\s*\$\s+(.+)$")

    def __init__(self, drop_keys=None):
        super().__init__()
        self.data = OrderedDict()
        if drop_keys is None:
            drop_keys = []
        self.drop_keys = set(drop_keys)

    def feed(self, line):
        """Process the given line

        See ``BaseParser.feed``
        """
        m = self.DIRECTIVE_RE.match(line)
        if m is None:
            return False

        key, value = _split_sge_directive(m.group(1))
        if key not in self.drop_keys:
            self.data[key] = (value, line)

        return True


def _translate_export_vars(value):
    if value == b"all":
        return b"-V"
    if value == b"none":
        return
    return b"-v %s" % value


def _translate_mail_type(value):
    trans = {b"none": b"n", b"begin": b"b", b"end": b"e", b"fail": b"a"}
    vals = value.split(b",")
    newvals = []
    for val in vals:
        newval = trans.get(val.lower())
        if newval is None:
            _logger.warn("Unknown mail_type value %r", val)
            newval = val
        newvals.append(newval)
    return b"-m %s" % b"".join(newvals)


class SGESite(Site):
    """Site managed using SGE"""

    directive_prefix = b"#$ "
    directive_translate = {
        "billing_account": b"-A %s",
        "error_file": b"-e %s",
        "export_vars": _translate_export_vars,
        "mail_type": _translate_mail_type,
        "join_output_error": b"-j y",  # TODO: make that automatic
        "mail_user": b"-M %s",
        "name": b"-N %s",
        "output_file": b"-o %s",
        "priority": b"-p %s",
        "queue": b"-q %s",
        "walltime": b"-l h_rt=%s",
    }

    SUBMIT_RE = re.compile(r"(Your job )?(\d+)")

    def __init__(self, config, connection, global_config):
        super().__init__(config, connection, global_config)
        self._qsub = command_as_list(config.get("qsub_command", "qsub"))
        self._qdel = command_as_list(config.get("qdel_command", "qdel"))
        self._qstat = command_as_list(config.get("qstat_command", "qstat"))
        self._copy_script = config.get("copy_script", False)
        self._copy_jid = config.get("copy_jid", False)

    def _parse_submit_output(self, out):
        match = self.SUBMIT_RE.search(out)
        if match is None:
            _logger.warn("Could not parse SGE output %r", out)
            return None
        return int(match.group(2))

    def submit(self, script, user, output, dryrun=False):
        """See `troika.sites.Site.submit`"""
        script = pathlib.Path(script)

        cmd = self._qsub.copy()

        if not script.exists():
            raise InvocationError(f"Script file {str(script)!r} does not exist")
        inpf = None
        if self._copy_script:
            script_remote = pathlib.PurePath(output).parent / script.name
            self._connection.sendfile(script, script_remote, dryrun=dryrun)
            cmd.append(script_remote)
        else:
            inpf = script.open(mode="rb")

        proc = self._connection.execute(
            cmd, stdin=inpf, stdout=PIPE, stderr=PIPE, dryrun=dryrun
        )
        if dryrun:
            return

        proc_stdout, proc_stderr = proc.communicate()
        if proc.returncode != 0:
            if proc_stdout:
                _logger.error(
                    "qsub stdout for script %s:\n%s", script, proc_stdout.strip()
                )
            if proc_stderr:
                _logger.error(
                    "qsub stderr for script %s:\n%s", script, proc_stderr.strip()
                )
            check_retcode(proc.returncode, what="Submission")
        else:
            if proc_stdout:
                _logger.debug(
                    "qsub stdout for script %s:\n%s", script, proc_stdout.strip()
                )
            if proc_stderr:
                _logger.debug(
                    "qsub stderr for script %s:\n%s", script, proc_stderr.strip()
                )

        jobid = self._parse_submit_output(
            proc_stdout.decode(locale.getpreferredencoding()).strip()
        )
        _logger.debug("SGE job ID: %s", jobid)

        jid_output = script.with_suffix(script.suffix + ".jid")
        if jid_output.exists():
            _logger.warning(
                "Job ID output file %r already exists, " + "overwriting",
                str(jid_output),
            )
        jid_output.write_text(str(jobid) + "\n")

        if self._copy_jid:
            jid_remote = pathlib.PurePath(output).parent / jid_output.name
            _logger.debug("Copying JID to output directory: %s", jid_remote)
            self._connection.sendfile(jid_output, jid_remote, dryrun=dryrun)

        return jobid

    def monitor(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.Site.monitor`"""
        script = pathlib.Path(script)

        if jid is None:
            jid = self._parse_jidfile(script, output)
            _logger.debug(f"Read job id {jid!r} from jidfile")
        else:
            _logger.debug(f"Using specified job id {jid!r}")

        stat_output = script.with_suffix(script.suffix + ".stat")
        if stat_output.exists():
            _logger.warning(
                "Status file %r already exists, overwriting", str(stat_output)
            )
        outf = None
        if not dryrun:
            outf = stat_output.open(mode="wb")

        self._connection.execute(self._qstat + ["-j", jid], stdout=outf, dryrun=dryrun)

        _logger.info("Output written to %r", str(stat_output))

    def kill(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.Site.kill`"""
        script = pathlib.Path(script)

        if jid is None:
            jid = self._parse_jidfile(script, output)
            _logger.debug(f"Read job id {jid!r} from jidfile")
        else:
            _logger.debug(f"Using specified job id {jid!r}")

        cmd = self._qdel + [jid]
        proc = self._connection.execute(cmd, stdout=PIPE, dryrun=dryrun)

        if dryrun:
            return (jid, "KILLED")

        proc_stdout, _ = proc.communicate()
        retcode = proc.returncode
        if retcode != 0:
            _logger.error("qdel output: %s", proc_stdout)
            check_retcode(retcode, what="Kill")

        return (jid, "KILLED")

    def get_native_parser(self):
        """See `troika.sites.Site.get_native_parser`"""
        return SGEDirectiveParser(drop_keys=[b"-o", b"-e", b"-j"])

    def _parse_jidfile(self, script, output=None, dryrun=False):
        script = pathlib.Path(script)
        jid_output = script.with_suffix(script.suffix + ".jid")
        try:
            return jid_output.read_text().strip()
        except IOError as e:
            if self._copy_jid and output is not None:
                jid_remote = pathlib.PurePath(output).parent / jid_output.name
                try:
                    self._connection.getfile(jid_remote, jid_output, dryrun=dryrun)
                    _logger.debug(
                        "Job ID file copied back from output directory: %s", jid_remote
                    )
                    if not dryrun:
                        return jid_output.read_text().strip()
                except (IOError, RunError) as e2:
                    raise RunError(
                        f"Could not read the job id: {e!s} or copy it back {e2!s}"
                    )
            raise RunError(f"Could not read the job id: {e!s}")

    def __repr__(self):
        return f"{self.__class__.__name__}(connection={self._connection!r}, qsub_command={self._qsub[0]!r})"



"""Common directive translators"""

from .. import ConfigurationError


def join_output_error(script_data, global_config, site):
    """Add the 'join_output_error' directive when needed"""
    if "error_file" not in script_data["directives"]:
        script_data["directives"]["join_output_error"] = ()
    return script_data


def extra_directives(script_data, global_config, site):
    """Add the content of 'extra_directives' in the site configuration"""
    for name, val in site.config.get("extra_directives", {}).items():
        if isinstance(val, (int, float)):
            val = str(val)
        if not isinstance(val, str):
            raise ConfigurationError(
                f"Invalid value type for directive {name!r}: {val!r}"
            )
        script_data["directives"].setdefault(name, val.encode("utf-8"))
    return script_data



"""Common definitions"""

VERSION = "0.2.6"


class ConfigurationError(RuntimeError):
    """Exception raised in case of an invalid configuration"""

    pass


class InvocationError(RuntimeError):
    """Exception raised in case of an invalid parameter"""

    pass


class RunError(RuntimeError):
    """Exception raised by runtime failures"""

    pass



"""Local connection class"""

import logging
import os
import pathlib
import shutil
import subprocess
from subprocess import DEVNULL, STDOUT

from .base import Connection

_logger = logging.getLogger(__name__)


class LocalConnection(Connection):
    """Connection to the local host"""

    def __init__(self, config, user):
        super().__init__(config, user)
        self.local_cwd = config.get("local_cwd", None)
        if self.local_cwd:
            self.local_cwd = pathlib.PurePath(self.local_cwd)

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def is_local(self):
        """See `Connection.is_local`"""
        return True

    def get_parent(self):
        """See `Connection.get_parent`"""
        return self

    def execute(
        self,
        command,
        stdin=None,
        stdout=None,
        stderr=None,
        text=False,
        encoding=None,
        errors=None,
        detach=False,
        env=None,
        cwd=None,
        dryrun=False,
    ):
        """See `Connection.execute`"""
        if dryrun:
            _logger.info("Execute: %s", " ".join(repr(str(arg)) for arg in command))
            return
        if stdin is None:
            stdin = DEVNULL
        if stdout is None:
            stdout = DEVNULL
        if stderr is None:
            stderr = STDOUT
        if cwd is None:
            cwd = self.local_cwd
        elif self.local_cwd is not None:
            # Treat cwd relative to default if present
            cwd = self.local_cwd / cwd
        _logger.debug("Executing %s", " ".join(repr(str(arg)) for arg in command))
        proc = subprocess.Popen(
            command,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            text=text,
            encoding=encoding,
            errors=errors,
            start_new_session=detach,
            env=({**os.environ, **env} if env is not None else None),
            cwd=cwd,
        )
        _logger.debug("Child PID: %d", proc.pid)
        return proc

    def sendfile(self, src, dst, dryrun=False):
        """See `Connection.sendfile`"""
        if self.local_cwd:
            # If dst is relative, treat it relative to configured cwd
            dst = self.local_cwd / dst
        # but src is always relative to Troika process, not connection
        if not dryrun:
            shutil.copy(src, dst)
        else:
            _logger.info("Copying %r to %r", src, dst)

    def getfile(self, src, dst, dryrun=False):
        """See `Connection.getfile`"""
        if self.local_cwd:
            # If src is relative, treat it relative to configured cwd
            src = self.local_cwd / src
        # but dst is always relative to Troika process, not connection
        if not dryrun:
            shutil.copy(src, dst)
        else:
            _logger.info("Copying %r to %r", src, dst)



"""Various utilities"""

import logging
import signal

from . import RunError

_logger = logging.getLogger(__name__)


def normalise_signal(sig):
    """Get the `signal.Signals` value associated with the given signal

    >>> normalise_signal(2)
    <Signals.SIGINT: 2>
    >>> normalise_signal('KILL')
    <Signals.SIGKILL: 9>
    >>> normalise_signal('SIGTERM')
    <Signals.SIGTERM: 15>
    >>> normalise_signal('NOTASIGNAL')
    Traceback (most recent call last):
        ...
    ValueError: Invalid signal: 'NOTASIGNAL'
    >>> normalise_signal(-3)
    Traceback (most recent call last):
        ...
    ValueError: Invalid signal: -3
    """
    if isinstance(sig, signal.Signals):
        return sig
    elif isinstance(sig, str):
        name = sig.upper()
        if not name.startswith("SIG"):
            name = "SIG" + name
        match = [s for s in signal.Signals if s.name == name]
        if len(match) == 1:
            return match[0]
    elif isinstance(sig, int):
        match = [s for s in signal.Signals if s.value == int(sig)]
        if len(match) == 1:
            return match[0]
    raise ValueError(f"Invalid signal: {sig!r}")


def check_retcode(retcode, what="Command", suffix=""):
    """Check a return code and raise an error if needed

    Parameters
    ----------
    retcode: int
        Return code, see `subprocess.Popen.returncode`
    what: str
        Description of what was run
    suffix: str
        Extra content to append to the error message

    Raises
    ------
    `troika.RunError`
        The return code was nonzero


    >>> check_retcode(0)
    >>> check_retcode(1)
    Traceback (most recent call last):
        ...
    troika.RunError: Command failed with exit code 1
    >>> check_retcode(-6)
    Traceback (most recent call last):
        ...
    troika.RunError: Command terminated by signal 6
    """
    if retcode != 0:
        msg = what
        if retcode > 0:
            msg += f" failed with exit code {retcode}"
        else:
            msg += f" terminated by signal {-retcode}"
        msg += suffix
        raise RunError(msg)


def first_not_none(lst):
    """Return the first element in `l` that is not None, if any

    >>> first_not_none(['a', 'b', 'c'])
    'a'
    >>> first_not_none([None, 1, 3])
    1
    >>> first_not_none([2, None, 6])
    2
    >>> first_not_none([None, None])
    >>> first_not_none([])
    """
    for x in lst:
        if x is not None:
            return x
    return None


_parse_bool_sentinel = object()


def parse_bool(x, default=_parse_bool_sentinel):
    """Parse a boolean value

    Parameters
    ----------
    x: any
        Input value. Booleans are returned unchanged. True strings are "yes",
        "1", "true", and "on" (case insensitive). False strings are "no", "0",
        "false", "off" (case insensitive). Integer value 0 means False, 1
        means True. Any other value is considered invalid.
    default: any
        Return value in case of an invalid input

    Returns
    -------
    bool or type(default)

    Raises
    ------
    `ValueError`
        The input is invalid and no default is provided


    >>> all(parse_bool(x) for x in [True, "yes", "Yes", "true", "True", "on", "ON", "1", 1])
    True
    >>> any(parse_bool(x) for x in [False, "no", "No", "false", "False", "off", "OFF", "0", 0])
    False
    >>> all(parse_bool(x) for x in [b"yes", b"Yes", b"true", b"True", b"on", b"ON", "1"])
    True
    >>> any(parse_bool(x) for x in [b"no", b"No", b"false", b"False", b"off", b"OFF", b"0"])
    False
    >>> parse_bool("maybe", default=False)
    False
    >>> parse_bool("maybe", default=True)
    True
    >>> parse_bool(None, default=False)
    False
    >>> parse_bool([], default=True)
    True
    >>> parse_bool(3, default=False)
    False
    >>> parse_bool("invalid")
    Traceback (most recent call last):
        ...
    ValueError: Cannot parse boolean 'invalid'
    """
    if isinstance(x, bool):
        return x
    if isinstance(x, bytes):
        x = x.decode("ascii", "replace")
    if isinstance(x, str):
        x = x.lower()
        if x in ["no", "0", "false", "off"]:
            return False
        elif x in ["yes", "1", "true", "on"]:
            return True
    elif isinstance(x, int):
        if x == 0:
            return False
        elif x == 1:
            return True

    if default is not _parse_bool_sentinel:
        return default

    raise ValueError(f"Cannot parse boolean {x!r}")


def command_as_list(x):
    """Return the given command as a list

    >>> command_as_list("foo")
    ['foo']
    >>> command_as_list("spaces not split")
    ['spaces not split']
    >>> command_as_list(b"bytes")
    [b'bytes']
    >>> command_as_list(["prg", "arg1", "arg2"])
    ['prg', 'arg1', 'arg2']
    """
    if isinstance(x, (str, bytes)):
        return [x]
    return list(x)



"""SSH connection class"""

import logging
import pathlib
import shlex

from ..connection import PIPE
from ..utils import check_retcode, parse_bool
from .base import Connection
from .local import LocalConnection

_logger = logging.getLogger(__name__)


class SSHConnection(Connection):
    """Connection to a remote host via SSH"""

    def __init__(self, config, user):
        super().__init__(config, user)
        self.parent = LocalConnection(config, user)
        self.ssh = config.get("ssh_command", "ssh")
        self.scp = config.get("scp_command", "scp")
        self.ssh_options = config.get("ssh_options", [])
        self.scp_options = config.get("scp_options", self.ssh_options.copy())
        if parse_bool(config.get("ssh_verbose", False)):
            self.ssh_options.append("-v")
            self.scp_options.append("-v")
        strict_host_key_checking = parse_bool(
            config.get("ssh_strict_host_key_checking", False)
        )
        if strict_host_key_checking is not None:
            self.ssh_options.append(
                f'-oStrictHostKeyChecking={"yes" if strict_host_key_checking else "no"}'
            )
            self.scp_options.append(
                f'-oStrictHostKeyChecking={"yes" if strict_host_key_checking else "no"}'
            )
        connect_timeout = config.get("ssh_connect_timeout", None)
        if connect_timeout is not None:
            self.ssh_options.append(f"-oConnectTimeout={connect_timeout}")
            self.scp_options.append(f"-oConnectTimeout={connect_timeout}")
        self.host = config["host"]
        if self.user is None:
            self.user = config.get("user", None)
        self.remote_cwd = config.get("remote_cwd", None)
        if self.remote_cwd:
            self.remote_cwd = pathlib.PurePath(self.remote_cwd)

    def __repr__(self):
        return f"{self.__class__.__name__}(host={self.host!r}, user={self.user!r})"

    def get_parent(self):
        """See `Connection.get_parent`"""
        return self.parent

    def execute(
        self,
        command,
        stdin=None,
        stdout=None,
        stderr=None,
        text=False,
        encoding=None,
        errors=None,
        detach=False,
        env=None,
        cwd=None,
        dryrun=False,
    ):
        """See `Connection.execute`"""
        args = [self.ssh] + self.ssh_options
        if self.user is None:
            args.append(f"{self.host}")
        else:
            args.append(f"{self.user}@{self.host}")
        if cwd is None:
            cwd = self.remote_cwd
        elif self.remote_cwd is not None:
            # Treat cwd relative to default if present
            cwd = self.remote_cwd / cwd
        if cwd is not None:
            args += ["cd", shlex.quote(str(cwd)), "&&"]
        if env is not None:
            args += [f"{shlex.quote(k)}={shlex.quote(v)}" for k, v in env.items()]
        args += [shlex.quote(str(arg)) for arg in command]
        return self.parent.execute(
            args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            text=text,
            encoding=encoding,
            errors=errors,
            detach=detach,
            dryrun=dryrun,
        )

    def sendfile(self, src, dst, dryrun=False):
        """See `Connection.sendfile`"""
        if self.parent.local_cwd is not None:
            # src is always relative to Troika process, not underlying LocalConnection
            src = pathlib.Path(src).absolute()
        if self.remote_cwd:
            # If dst is relative, treat it relative to configured cwd
            dst = self.remote_cwd / dst
        scp_args = [self.scp] + self.scp_options + [src]
        if self.user is None:
            scp_args.append(f"{self.host}:{dst}")
        else:
            scp_args.append(f"{self.user}@{self.host}:{dst}")
        proc = self.parent.execute(scp_args, stdout=PIPE, stderr=PIPE, dryrun=dryrun)
        if dryrun:
            return
        proc_stdout, proc_stderr = proc.communicate()
        proc_stdout = proc_stdout.strip()
        proc_stderr = proc_stderr.strip()
        retcode = proc.wait()
        if proc_stdout:
            _logger.debug("scp output: %s", proc_stdout)
        if retcode != 0:
            if proc_stderr:
                _logger.error("scp error: %s", proc_stderr)
        else:
            if proc_stderr:
                _logger.debug("scp error output: %s", proc_stderr)
        check_retcode(retcode, what="Copy")

    def getfile(self, src, dst, dryrun=False):
        """See `Connection.getfile`"""
        if self.remote_cwd:
            # If src is relative, treat it relative to configured cwd
            src = self.remote_cwd / src
        if self.parent.local_cwd is not None:
            # dst is always relative to Troika process, not underlying LocalConnection
            dst = pathlib.Path(dst).absolute()
        scp_args = [self.scp] + self.scp_options
        if self.user is None:
            scp_args.append(f"{self.host}:{src}")
        else:
            scp_args.append(f"{self.user}@{self.host}:{src}")
        scp_args.append(dst)
        proc = self.parent.execute(scp_args, stdout=PIPE, stderr=PIPE, dryrun=dryrun)
        if dryrun:
            return
        proc_stdout, proc_stderr = proc.communicate()
        proc_stdout = proc_stdout.strip()
        proc_stderr = proc_stderr.strip()
        retcode = proc.wait()
        if proc_stdout:
            _logger.debug("scp output: %s", proc_stdout)
        if retcode != 0:
            if proc_stderr:
                _logger.error("scp error: %s", proc_stderr)
        else:
            if proc_stderr:
                _logger.debug("scp error output: %s", proc_stderr)
        check_retcode(retcode, what="Copy")



"""Directive translator"""

import logging

from ..hooks.base import Hook

_logger = logging.getLogger(__name__)


class Translator(Hook):
    """Directive translation hook manager

    See :py:class:`troika.hooks.base.Hook`. The only change is the behaviour of the
    :py:meth:`__call__` method::

        def __call__(self, script_data, *args, **kwargs):
            for func in self.enabled_hooks:
                script_data = func(script_data, *args, **kwargs)
            return script_data
    """

    _namespace = "troika"

    def __call__(self, data, *args, **kwargs):
        if self._impl is None:
            raise ValueError("Attempt to call a non-instantiated hook registry")
        _logger.debug("Translating directives")
        for funcname, func in self._impl:
            _logger.debug("Calling translator function %s", funcname)
            data = func(data, *args, **kwargs)
        return data


@Translator.declare
def translators(script_data, global_config, site):
    """Directive translation hook

    Parameters
    ----------
    script_data: dict[str, Any]
        Data parsed from the input script
    global_config: :py:class:`troika.config.Config`
        Global configuration
    site: :py:class:`troika.sites.base.Site`
        Target site

    Returns
    -------
    dict[str, Any]
        Updated ``script_data``
    """



"""Abstraction of the way to connect to a host"""

import logging
from subprocess import DEVNULL, PIPE, STDOUT  # noqa

from . import ConfigurationError
from .components import get_entrypoint

_logger = logging.getLogger(__name__)


def get_connection(name, config, user):
    """Load the requested `troika.connections.base.Connection` object"""

    try:
        cls = get_entrypoint("troika.connections", name)
    except ValueError:
        raise ConfigurationError(f"Unknown connection {name!r}")

    conn = cls(config, user)
    _logger.debug("Created connection %r", conn)
    return conn



"""ecFlow hooks"""

import logging
import pathlib

from .. import InvocationError, RunError
from ..connection import PIPE
from ..connections.local import LocalConnection
from ..parser import DirectiveParser
from ..utils import check_retcode

_logger = logging.getLogger(__name__)


def abort_on_ecflow(site, script, output, jid, cancel_status, dryrun=False):
    """Post-kill hook to issue an abort on behalf of a job that was killed
    or cancelled without the opportunity to inform ecFlow itself."""
    if cancel_status == "CANCELLED":
        msg = "Cancelled before starting"
    elif cancel_status == "KILLED":
        msg = "Killed forcefully"
    elif cancel_status == "VANISHED":
        msg = "Vanished unexpectedly"
    elif cancel_status == "TERMINATED":
        return
    else:
        raise InvocationError(
            f'abort_on_ecflow: unknown cancel status "{cancel_status}"'
        )

    script = pathlib.Path(script)
    orig_script = script.with_suffix(script.suffix + ".orig")

    if not orig_script.exists():
        orig_script_copy = pathlib.PurePath(output).parent / orig_script.name
        if output is not None:
            try:
                site._connection.getfile(orig_script_copy, orig_script, dryrun=dryrun)
                _logger.debug(
                    f"Original script copied back from output directory: {orig_script_copy!r}"
                )
            except (IOError, RunError) as e:
                raise RunError(f"Could not copy back original script {e!s}")

    parser = DirectiveParser()
    with open(orig_script, "rb") as sin:
        for line in sin:
            parser.feed(line)
    env = {}
    for directive, var, required in [
        ("ecflow_name", "ECF_NAME", True),
        ("ecflow_pass", "ECF_PASS", True),
        ("ecflow_host", "ECF_HOST", False),
        ("ecflow_port", "ECF_PORT", False),
    ]:
        try:
            env[var] = parser.data[directive].decode("ascii")
        except KeyError:
            if required:
                _logger.error(
                    f"abort_on_ecflow could not find {directive} defined in script {script}"
                )
                raise

    cmd = [parser.data.get("ecflow_client", "ecflow_client"), f"--abort={msg}"]

    if site._connection.is_local():
        _logger.debug(
            f'abort_on_ecflow running {" ".join(cmd)} on local site with env {env!r}'
        )
        connection = site._connection
    elif "ecflow_host" in env:
        _logger.debug(
            f'abort_on_ecflow running {" ".join(cmd)} on remote site with env {env!r}'
        )
        connection = site._connection
    else:
        _logger.debug(
            f'abort_on_ecflow running {" ".join(cmd)} locally with env {env!r}'
        )
        connection = LocalConnection({}, site._connection.user)

    proc = connection.execute(cmd, stdout=PIPE, stderr=PIPE, env=env, dryrun=dryrun)
    if dryrun:
        return

    proc_stdout, proc_stderr = proc.communicate()
    if proc.returncode != 0:
        if proc_stdout:
            _logger.error(
                "ecflow_client stdout for script %s:\n%s", script, proc_stdout.strip()
            )
        if proc_stderr:
            _logger.error(
                "ecflow_client stderr for script %s:\n%s", script, proc_stderr.strip()
            )
        check_retcode(proc.returncode, what="Abort")
    else:
        if proc_stdout:
            _logger.debug(
                "ecflow_client stdout for script %s:\n%s", script, proc_stdout.strip()
            )
        if proc_stderr:
            _logger.debug(
                "ecflow_client stderr for script %s:\n%s", script, proc_stderr.strip()
            )



"""Site group class"""

import logging

from .. import RunError
from ..site import get_site
from .base import Site

_logger = logging.getLogger(__name__)


class SiteGroup(Site):
    """Site group: choose the first site available"""

    __type_name__ = "group"

    def __init__(self, config, connection, global_config):
        self._select(config, connection.user, global_config)
        self._connection = self._selected._connection

    def _select(self, config, user, global_config):
        """Find a suitable site"""
        sites = config.get("sites", [])
        self._selected = None
        for name in sites:
            _logger.debug("Trying site %r", name)
            site = get_site(global_config, name, user)
            if self._check(site):
                self._selected = site
                break
        if self._selected is None:
            raise RunError("No site available in the group")

    def _check(self, site):
        """Check whether a given site is suitable"""
        return site.check_connection()

    def preprocess(self, script, user, output):
        """See `troika.sites.base.Site.preprocess`"""
        return self._selected.preprocess(script, user, output)

    def submit(self, script, user, output, dryrun=False):
        """See `troika.sites.base.Site.submit`"""
        return self._selected.submit(script, user, output, dryrun=dryrun)

    def monitor(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.base.Site.monitor`"""
        return self._selected.monitor(
            script, user, output=output, jid=jid, dryrun=dryrun
        )

    def kill(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.base.Site.kill`"""
        return self._selected.kill(script, user, output=output, jid=jid, dryrun=dryrun)

    def check_connection(self, timeout=None, dryrun=False):
        """See `troika.sites.base.Site.check_connection`"""
        return self._selected.check_connection(timeout=timeout, dryrun=dryrun)



"""Base connection class"""

import logging

from ..connection import PIPE

_logger = logging.getLogger(__name__)


class Connection:
    """Base connection class

    Parameters
    ----------
    config: dict
        Connection configuration
    """

    #: Value for the 'connection' key in the site configuration.
    #: If None, the name will be computed by turning the class name to
    #: lowercase and removing a trailing "connection" if present, e.g.
    #: ``FooConnection`` becomes ``foo``.
    __type_name__ = None

    def __init__(self, config, user):
        self.user = user

    def is_local(self):
        """Check whether the connection is local

        If the connection is local, local paths are valid through the connection
        """
        return False

    def get_parent(self):
        """Get the parent connection

        The parent connection can be used to interact with processes that have
        been created through the current connection.

        Returns
        -------
        Connection
            The parent connection
        """
        raise NotImplementedError

    def execute(
        self,
        command,
        stdin=None,
        stdout=None,
        stderr=None,
        text=False,
        encoding=None,
        errors=None,
        detach=False,
        env=None,
        cwd=None,
        dryrun=False,
    ):
        """Execute the given command on the host

        Parameters
        ----------
        command: list of str or path-like
            Command to execute, as a list of arguments
        stdin: None, PIPE or file-like
            Standard input, /dev/null if None
        stdout: None, PIPE or file-like
            Standard output, /dev/null if None
        stderr: None, PIPE, DEVNULL or file-like
            Standard error, same as stdout if None
        text: bool
            If True, open streams in text mode
        encoding: str or None
            Encoding to use for opening streams in text mode
        errors: str or None
            Error handling mode to use for decoding streams in text mode
        detach: bool
            If True, detach from the running command
        env: dict or None
            Extra variables to set in the command's environment
        cwd: path-like or None
            Override default working directory if not None
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed

        Returns
        -------
        :py:class:`subprocess.Popen` object or None
            Local process object associated to the connection, if dryrun is False,
            else None
        """
        raise NotImplementedError

    def sendfile(self, src, dst, dryrun=False):
        """Copy the given file to the remote host

        Parameters
        ----------
        src: path-like
            Path to the file on the local host
        dst: path-like
            Path to the target directory or file on the remote host
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed
        """
        raise NotImplementedError

    def getfile(self, src, dst, dryrun=False):
        """Get the given file from the remote host

        Parameters
        ----------
        src: path-like
            Path to the file on the remote host
        dst: path-like
            Path to the target directory or file on the local host
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed
        """
        raise NotImplementedError

    def checkstatus(self, timeout=None, dryrun=False):
        """Check whether the connection is working

        Parameters
        ----------
        timeout: int
            If set, consider the connection is not working if no response after
            this number of seconds
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed

        Returns
        -------
        bool
            True if the connection is able to execute commands
        """
        proc = self.execute(
            ["true"], stdout=PIPE, stderr=PIPE, detach=False, dryrun=dryrun
        )
        if dryrun:
            return True
        proc_stdout, proc_stderr = proc.communicate()
        retcode = proc.returncode
        if proc.returncode == 0:
            log = _logger.debug
        else:
            log = _logger.error
        if proc_stdout:
            log("stdout checking connection:\n%s", proc_stdout.strip())
        if proc_stderr:
            log("stderr checking connection:\n%s", proc_stderr.strip())
        return retcode == 0



"""Direct execution site"""

import logging
import os
import pathlib
import signal
import time

from .. import ConfigurationError, InvocationError, RunError
from .base import Site

_logger = logging.getLogger(__name__)


class DirectExecSite(Site):
    """Site where jobs are run directly"""

    __type_name__ = "direct"

    def __init__(self, config, connection, global_config):
        super().__init__(config, connection, global_config)
        self._copy_script = config.get("copy_script", False)
        self._copy_jid = config.get("copy_jid", False)
        self._shell = config.get(
            "shell", ["bash"] if self._copy_script else ["bash", "-s"]
        )
        self._use_shell = config.get("use_shell", not connection.is_local())

        if not (connection.is_local() or self._copy_script or self._use_shell):
            raise ConfigurationError(
                "copy_script and use_shell cannot both be False for a remote site"
            )

    def submit(self, script, user, output, dryrun=False):
        """See `troika.sites.base.Site.submit`"""
        script = pathlib.Path(script).resolve()
        if not script.exists():
            raise InvocationError(f"Script file {str(script)!r} does not exist")

        script_remote = script
        if self._copy_script and not self._connection.is_local():
            script_remote = pathlib.PurePath(output).parent / script.name
            super().create_output_dir(script_remote, dryrun=dryrun)
            self._connection.sendfile(script, script_remote, dryrun=dryrun)

        args = []
        if self._use_shell:
            args.extend(self._shell)
        if self._copy_script or (self._connection.is_local() and not self._use_shell):
            args.append(script_remote)

        inpf = None
        if self._use_shell and not self._copy_script:
            inpf = script.open(mode="rb")

        output = pathlib.Path(output)
        self.create_output_dir(output, dryrun=dryrun)
        if output.exists():
            _logger.warning("Output file %r already exists, overwriting", str(output))
        outf = None
        if not dryrun:
            outf = output.open(mode="wb")
        proc = self._connection.execute(
            args, stdin=inpf, stdout=outf, detach=True, dryrun=dryrun
        )

        if dryrun:
            return

        jid_output = script.with_suffix(script.suffix + ".jid")
        if jid_output.exists():
            _logger.warning(
                "Job ID output file %r already exists, " + "overwriting",
                str(jid_output),
            )
        jid_output.write_text(str(proc.pid) + "\n")

        if self._copy_jid:
            jid_remote = pathlib.PurePath(output).parent / jid_output.name
            _logger.debug("Copying JID to output directory: %s", jid_remote)
            self._connection.sendfile(jid_output, jid_remote, dryrun=dryrun)

        return proc

    def monitor(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.base.Site.monitor`"""
        script = pathlib.Path(script)

        if jid is None:
            jid = self._parse_jidfile(script, output)
            _logger.debug(f"Read job id {jid!r} from jidfile")
        else:
            _logger.debug(f"Using specified job id {jid!r}")
        try:
            jid = int(jid)
        except ValueError:
            raise RunError(f"Invalid job id: {jid!r}")

        stat_output = script.with_suffix(script.suffix + ".stat")
        if stat_output.exists():
            _logger.warning(
                "Status file %r already exists, overwriting", str(stat_output)
            )
        outf = None
        if not dryrun:
            outf = stat_output.open(mode="wb")

        conn = self._connection.get_parent()
        conn.execute(["ps", "-lyfp", str(jid)], stdout=outf, dryrun=dryrun)

        _logger.info("Output written to %r", str(stat_output))

    def kill(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.base.Site.kill`"""
        script = pathlib.Path(script)

        if jid is None:
            jid = self._parse_jidfile(script, output)
            _logger.debug(f"Read job id {jid!r} from jidfile")
        else:
            _logger.debug(f"Using specified job id {jid!r}")
        try:
            jid = int(jid)
        except ValueError:
            raise RunError(f"Invalid job id: {jid!r}")

        seq = self._kill_sequence
        if not seq:
            seq = [(0, signal.SIGTERM)]

        cancel_status = None
        for wait, sig in seq:
            time.sleep(wait)
            if sig is None:
                sig = signal.SIGTERM

            if dryrun:
                _logger.info(f"Sending {sig.name} to process {jid}")
                continue

            _logger.debug(f"Sending {sig.name} to process {jid}")
            try:
                os.kill(jid, sig.value)
            except ProcessLookupError:
                if cancel_status is None:
                    raise RunError(f"Process ID {jid} not found")
                else:
                    break

            if sig == signal.SIGKILL:
                cancel_status = "KILLED"
            else:
                cancel_status = "TERMINATED"

        return (jid, cancel_status)

    def create_output_dir(self, output, dryrun=False):
        """See `troika.sites.base.Site.create_output_dir`"""
        out_dir = pathlib.Path(output).parent
        if dryrun:
            _logger.info("Ensuring directory %r exists", str(out_dir))
        else:
            _logger.debug("Ensuring directory %r exists", str(out_dir))
            out_dir.mkdir(parents=True, exist_ok=True)
        return out_dir

    def _parse_jidfile(self, script, output=None, dryrun=False):
        script = pathlib.Path(script)
        jid_output = script.with_suffix(script.suffix + ".jid")
        try:
            return jid_output.read_text().strip()
        except IOError as e:
            if self._copy_jid and output is not None:
                jid_remote = pathlib.PurePath(output).parent / jid_output.name
                try:
                    self._connection.getfile(jid_remote, jid_output, dryrun=dryrun)
                    _logger.debug(
                        "Job ID file copied back from output directory: %s", jid_remote
                    )
                    if not dryrun:
                        return jid_output.read_text().strip()
                except (IOError, RunError) as e2:
                    raise RunError(
                        f"Could not read the job id: {e!s} or copy it back {e2!s}"
                    )
            raise RunError(f"Could not read the job id: {e!s}")

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"connection={self._connection!r}, "
            f"use_shell={self._use_shell}, "
            f"shell={self._shell!r})"
        )



"""Command-line interface"""

import argparse
import logging
import pathlib
import sys
import textwrap

from . import VERSION, ConfigurationError, InvocationError, RunError, log
from .config import get_config
from .controller import get_controller

_logger = logging.getLogger(__name__)

_config_guesses = [
    p / "etc" / "troika.yml"
    for p in [
        pathlib.Path(__file__).parent.parent.parent.parent.parent,
        pathlib.Path(sys.prefix),
    ]
]


class Action:
    """Command-line action

    Parameters
    ----------
    args: `argparse.Namespace`-like
    """

    #: If True, write the logs to a file
    save_log = True

    def __init__(self, args):
        self.logfile = None
        if self.save_log:
            self.logfile = log.get_logfile_path(
                args.action, getattr(args, "script", None)
            )
        if args.logfile is not None:
            self.logfile = args.logfile
        self.logmode = "a" if args.append_log else "w"
        log.config(args.verbose - args.quiet, self.logfile, self.logmode)
        self.args = args

    def execute(self):
        """Execute the action"""
        try:
            config = get_config(self.args.config, guesses=_config_guesses)
            controller = get_controller(config, self.args, self.logfile)
            return self.run(config, controller)
        except ConfigurationError as e:
            _logger.critical("Configuration error: %s", e)
            return 1
        except InvocationError as e:
            _logger.critical("Invocation error: %s", e)
            return 1
        except RunError as e:
            _logger.critical("%s", e)
            return 1
        except Exception:
            _logger.exception("Unhandled exception")
            return 1

    def run(self, config, controller):
        """Main entry point for the action

        Parameters
        ----------
        config: `troika.config.Config`
        controller: `troika.controller.Controller`

        Returns
        -------
        int
            Exit code
        """
        raise NotImplementedError


class SubmitAction(Action):
    """Main entry point for the 'submit' sub-command"""

    def run(self, config, controller):
        args = self.args
        return controller.submit(args.script, args.user, args.output, args.dryrun)


class MonitorAction(Action):
    """Main entry point for the 'monitor' sub-command"""

    def run(self, config, controller):
        args = self.args
        return controller.monitor(
            args.script, args.user, args.output, args.jobid, args.dryrun
        )


class KillAction(Action):
    """Main entry point for the 'kill' sub-command"""

    def run(self, config, controller):
        args = self.args
        return controller.kill(
            args.script, args.user, args.output, args.jobid, args.dryrun
        )


class CheckConnectionAction(Action):
    """Main entry point for the 'check-connection' sub-command"""

    save_log = False

    def run(self, config, controller):
        args = self.args
        working = controller.check_connection(args.timeout, args.dryrun)
        if working:
            print("OK")
            return 0
        print("Connection failed", file=sys.stderr)
        return 1


class ListSitesAction(Action):
    """Main entry point for the 'list-sites' sub-command"""

    save_log = False

    def run(self, config, controller):
        print("Available sites:")
        print(
            "{name:<28s} {tp:<15s} {conn:<15s}".format(
                name="Name", tp="Type", conn="Connection"
            )
        )
        print("-" * 60)
        for name, tp, conn in controller.list_sites():
            print(f"{name:<28s} {tp:<15s} {conn:<15s}")
        return 0


def main(args=None, prog=None):
    """Main entry point

    Parameters
    ----------
    args: None or list of str
        Use these command-line arguments instead of sys.argv
    prog: None or str
        Use this program name

    Returns
    -------
    int
        Exit code
    """

    epilog = textwrap.dedent(
        """\
        environment variables:
          TROIKA_CONFIG_FILE    path to the default configuration file
    """
    )

    parser = argparse.ArgumentParser(
        prog=prog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Submit, monitor and kill jobs on remote systems",
        epilog=epilog,
    )

    parser.add_argument(
        "-V", "--version", action="version", version=("%(prog)s " + VERSION)
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase verbosity level (can be repeated)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="decrease verbosity level (can be repeated)",
    )
    parser.add_argument(
        "-l", "--logfile", default=None, help="save log output to this file"
    )
    parser.add_argument(
        "-A",
        "--append-log",
        default=False,
        action="store_true",
        help="append to the log file instead of overwriting",
    )

    parser.add_argument(
        "-c",
        "--config",
        type=argparse.FileType("r"),
        default=None,
        help="path to the configuration file",
    )
    parser.add_argument(
        "-n",
        "--dryrun",
        default=False,
        action="store_true",
        help="if true, do not execute, just report",
    )

    subparsers = parser.add_subparsers(
        dest="action",
        metavar="action",
        help="perform this action, see `%(prog)s <action> --help` for details",
    )

    parser_submit = subparsers.add_parser("submit", help="submit a new job")
    parser_submit.set_defaults(act=SubmitAction)
    parser_submit.add_argument("site", help="target site")
    parser_submit.add_argument("script", help="job script")
    parser_submit.add_argument("-u", "--user", default=None, help="remote user")
    parser_submit.add_argument("-o", "--output", required=True, help="job output file")
    parser_submit.add_argument(
        "-D",
        "--define",
        default=[],
        action="append",
        metavar="NAME=VALUE",
        help="set these directives in the submitted job",
    )

    parser_monitor = subparsers.add_parser("monitor", help="monitor a submitted job")
    parser_monitor.set_defaults(act=MonitorAction)
    parser_monitor.add_argument("site", help="target site")
    parser_monitor.add_argument("script", help="job script")
    parser_monitor.add_argument("-u", "--user", default=None, help="remote user")
    parser_monitor.add_argument(
        "-o", "--output", required=False, help="job output file"
    )
    parser_monitor.add_argument(
        "-j",
        "--jobid",
        default=None,
        type=lambda j: None if j == "" else j,
        help="remote job ID",
    )

    parser_kill = subparsers.add_parser("kill", help="kill a submitted job")
    parser_kill.set_defaults(act=KillAction)
    parser_kill.add_argument("site", help="target site")
    parser_kill.add_argument("script", help="job script")
    parser_kill.add_argument("-u", "--user", default=None, help="remote user")
    parser_kill.add_argument("-o", "--output", required=False, help="job output file")
    parser_kill.add_argument(
        "-j",
        "--jobid",
        default=None,
        type=lambda j: None if j == "" else j,
        help="remote job ID",
    )

    parser_checkconn = subparsers.add_parser(
        "check-connection", help="check whether the connection works"
    )
    parser_checkconn.set_defaults(act=CheckConnectionAction)
    parser_checkconn.add_argument("site", help="target site")
    parser_checkconn.add_argument("-u", "--user", default=None, help="remote user")
    parser_checkconn.add_argument(
        "-t",
        "--timeout",
        default=None,
        type=int,
        help="wait at most this number of seconds",
    )

    parser_listsites = subparsers.add_parser("list-sites", help="list available sites")
    parser_listsites.set_defaults(act=ListSitesAction)

    args = parser.parse_args(args)

    if not hasattr(args, "act"):
        parser.error("please specify an action")

    action = args.act(args)
    return action.execute()



"""Hook system to perform custom actions"""

import logging

from .hooks.base import Hook, at_exit, at_startup, post_kill, pre_submit  # noqa

_logger = logging.getLogger(__name__)


def setup_hooks(config, site):
    """Set up the hooks according to site configuration

    Parameters
    ----------
    config: `troika.config.Config`
        Configuration object
    site: str
        Site name
    """
    site_config = config.get_site_config(site)
    for spec in Hook.registered_hooks.values():
        req = site_config.get(spec.name, [])
        spec.instantiate(req)
        if req:
            _logger.debug("Enabled %s hooks: %s", spec.name, ", ".join(req))



"""Logging system"""

import logging
import pathlib

LOGLEVELS = [
    logging.CRITICAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
]

_logger = logging.getLogger(__name__)


def get_logfile_path(action, script=None):
    """Construct the path to the default log file

    Parameters
    ----------
    action: str
        Action to perform (`'submit'`, `'monitor'`, `'kill'`)
    script: path-like
        Path to the job script

    Returns
    -------
    `pathlib.Path`
        Path to the log file: `<script>.<action>log`

    >>> str(get_logfile_path('submit', 'foo.sh'))
    'foo.sh.submitlog'
    >>> str(get_logfile_path('monitor', 'bar.sh'))
    'bar.sh.monitorlog'
    >>> str(get_logfile_path('kill', 'spam'))
    'spam.killlog'
    >>> str(get_logfile_path('submit'))
    'troika.submitlog'
    """

    base = pathlib.Path(script if script is not None else "troika")
    return base.with_suffix(base.suffix + f".{action}log")


def config(verbose=0, logfile=None, logmode="a"):
    """Configure logging

    Parameters
    ----------
    verbose: int
        Increase (or decrease) the default log level this many times
    logfile: path-like or None
        Path to a log file
    logmode: 'a' or 'w'
        Mode to use for the log file ([a]ppend or over[w]rite)
    """

    default_log = LOGLEVELS.index(logging.WARNING)
    loglevel = LOGLEVELS[max(0, min(len(LOGLEVELS) - 1, default_log + verbose))]

    log_format = "%(asctime)s; %(name)s; %(levelname)s; %(message)s"

    logging.basicConfig(format=log_format, level=loglevel)

    if logfile is not None:
        root_logger = logging.getLogger()
        try:
            fh = logging.FileHandler(logfile, mode=logmode)
        except IOError as e:
            _logger.error("Cannot open log file: %s", e)
            return

        fh.setFormatter(logging.Formatter(log_format))
        fh.setLevel(logging.DEBUG)
        root_logger.addHandler(fh)
        _logger.debug("Writing logs to %r", str(logfile))



"""Directive handling"""

from .translators.base import translators  # noqa

#: Directive aliases. A directive named `dir` found in the script will be
#: converted to `ALIASES[dir]`, if present
ALIASES = {
    "error": "error_file",
    "job_name": "name",
    "output": "output_file",
    "time": "walltime",
}



"""Script parsing logic

Troika directives are of the form::

    # troika key=value

To be recognised, a line **must** start with the # sign, optionally followed by
white space, the string "troika" (case-insensitive), at least one white space
character, and a key-value pair. Valid keys consist only of letters, numbers
and underscores, and do not start with a number. Whitespace is removed around
the equals sign and at the end the value.
"""

import re
from collections import OrderedDict

from . import InvocationError, RunError


class ParseError(RunError):
    """Exception raised during script parsing"""

    pass


class BaseParser:
    """Base parser class

    Members
    -------
    data: object
        Output data
    """

    data = None

    def __init__(self):
        pass

    def feed(self, line):
        """Feed a line to be processed

        Parameters
        ----------
        line: bytes
            Line to process

        Returns
        -------
        bool
            True if the line can be dropped from the script body
        """
        raise NotImplementedError


class DirectiveParser(BaseParser):
    """Parser that processes a script to extract directives

    Usage
    -----
    Parse a script::

        script = open("myscript", "rb")
        parser = Parser()
        for line in script:
            parser.feed(line)
        data = parser.data

    Members
    -------
    data: collections.OrderedDict[str, bytes]
        Directives that have been parsed
    """

    DIRECTIVE_RE = re.compile(rb"^#\s*troika\s+(.+?)\s*$", re.I)
    KEYVAL_RE = re.compile(rb"^([A-Za-z_][A-Za-z0-9_]*)\s*=\s*(.*)$")

    def __init__(self, aliases=None):
        super().__init__()
        self.data = OrderedDict()
        self.aliases = aliases if aliases is not None else {}

    def feed(self, line):
        """Process the given line

        See ``BaseParser.feed``
        """
        dm = self.DIRECTIVE_RE.match(line)
        if dm is None:
            return False

        kv = dm.group(1)
        kvm = self.KEYVAL_RE.match(kv)
        if kvm is None:
            raise ParseError(f"Invalid key-value pair: {kv}")

        key, value = kvm.groups()
        key = key.decode("ascii")
        key = self.aliases.get(key, key)
        self.data[key] = value

        return True

    def parse_directive_args(self, args):
        """Process a list of name=value arguments. Returns a dict."""
        data = {}
        for arg in args:
            arg = arg.encode("utf-8")
            m = self.KEYVAL_RE.match(arg)
            if m is None:
                raise InvocationError(f"Invalid key-value pair: {arg!r}")
            key, value = m.groups()
            key = key.decode("ascii")
            key = self.aliases.get(key, key)
            data[key] = value
        return data


class ShebangParser(BaseParser):
    """Parser that extract the 'shebang' line

    The 'shebang' line, if present, must start with '#!' and be the first
    non-blank line fed to the parser.

    Members
    -------
    data: bytes or None
        Shebang line, if found, None otherwise
    """

    def __init__(self):
        super().__init__()
        self.done = False

    def feed(self, line):
        """Process the given line

        See ``BaseParser.feed``
        """
        if self.done:
            return False
        if line.isspace():
            return False
        self.done = True
        if line.startswith(b"#!"):
            self.data = line
            return True
        return False


class MultiParser(BaseParser):
    """Composition of multiple parsers

    A line fed to this parser is fed to each subparser in order, until the
    first True return value.

    Parameters
    ----------
    parsers: [(str, BaseParser)]
        Sub-parsers with labels
    """

    def __init__(self, parsers):
        super().__init__()
        self.parsers = parsers

    def feed(self, line):
        """Process the given line

        See ``BaseParser.feed``
        """
        for _, parser in self.parsers:
            drop = parser.feed(line)
            if drop:
                return True
        return False

    @property
    def data(self):
        return {label: parser.data for label, parser in self.parsers}



"""Script header generation logic"""

import logging

from . import ConfigurationError, InvocationError

_logger = logging.getLogger(__name__)


def ignore(value):
    """Ignore the requested directive"""
    return None


class Generator:
    """Base script header generator

    Parameters
    ----------
    directive_prefix: bytes or None
        Prefix for the generated directives, e.g. b"#SBATCH ". If `None`, no
        directives will be generated

    directive_translate: dict[str, bytes]
        Directive translation table. Values can be either `bytes` that are
        formatted using the % operator, or callable objects that are called with
        the value of the directive and should return `None` (ignore the
        directive), a byte string, or a list of byte strings

    unknown_directive: str, optional (`'fail'`, `'warn'`, or `'ignore'`)
        If set to `'fail'`, an unknown directive will cause Troika to exit with
        an error. If `'warn'` (the default), a warning will be shown and
        execution will continue. If `'ignore'`, execution will continue without
        warning.
    """

    def __init__(self, directive_prefix, directive_translate, unknown_directive="warn"):
        self.dir_prefix = directive_prefix
        self.dir_translate = directive_translate
        if unknown_directive not in ("fail", "warn", "ignore"):
            raise ConfigurationError(
                f"Invalid unknown directive behaviour: {unknown_directive!r},"
                + "should be 'fail', 'warn', or 'ignore'"
            )
        self.unknown = unknown_directive

    def generate(self, script_data):
        """Generate the script header

        Parameters
        ----------
        script_data: dict
            Script data collected at the parsing stage

        Returns
        -------
        list
            Lines of the script header, with endings
        """

        header = []

        shebang = script_data.get("shebang")
        if shebang is not None:
            if not shebang.endswith(b"\n"):
                shebang += b"\n"
            header.append(shebang)

        if self.dir_prefix is not None:
            for name, arg in script_data["directives"].items():
                fmt = self.dir_translate.get(name)
                if fmt is None:
                    self._unknown_directive(name)
                    continue
                directives = None
                if isinstance(fmt, bytes):
                    directives = [fmt % arg]
                else:
                    directives = fmt(arg)
                    if directives is None:
                        directives = []
                    elif isinstance(directives, bytes):
                        directives = [directives]
                header.extend(
                    self.dir_prefix + directive + b"\n" for directive in directives
                )

        native = script_data.get("native")
        if native is not None:
            for _, directive in native.values():
                header.append(directive)

        extra = script_data.get("extra")
        if extra is not None:
            header.append(b"\n")
            header.extend(extra)

        return header

    def _unknown_directive(self, name):
        if self.unknown == "fail":
            raise InvocationError(f"Unknown directive {name!r}")
        if self.unknown == "warn":
            _logger.warning("Unknown directive %r", name)



"""Translators that operate on resources"""


def enable_hyperthreading(script_data, global_config, site):
    """Add the 'enable_hyperthreading' directive when needed"""
    enable = int(script_data["directives"].get("threads_per_core", 1)) > 1
    script_data["directives"].setdefault("enable_hyperthreading", enable)
    return script_data



"""Sites handling"""

import logging

from . import ConfigurationError, InvocationError
from .components import get_entrypoint
from .connection import get_connection

_logger = logging.getLogger(__name__)


def get_site(config, name, user):
    """Create a `troika.site.Site` object from configuration"""

    try:
        site_config = config.get_site_config(name)
    except KeyError:
        raise InvocationError(f"Unknown site {name!r}")

    try:
        tp = site_config["type"]
    except KeyError:
        raise ConfigurationError(f"Site {name!r} has no 'type'")

    try:
        cls = get_entrypoint("troika.sites", tp)
    except ValueError:
        raise ConfigurationError(f"Site {name!r} has unknown type {tp!r}")

    try:
        conn_name = site_config["connection"]
    except KeyError:
        raise ConfigurationError(f"Site {name!r} has no 'connection'")

    conn = get_connection(conn_name, site_config, user)
    site = cls(site_config, conn, config)
    _logger.debug("Created %r for %s", site, name)
    return site


def list_sites(config):
    """List available sites

    Parameters
    ----------
    config: `troika.config.Config`
        Configuration object

    Yields
    -------
    ``(name, type, connection)``
    """
    try:
        sites = config["sites"]
    except KeyError:
        raise ConfigurationError("No 'sites' defined in configuration")

    for name, site in sites.items():
        try:
            tp = site["type"]
        except KeyError:
            raise ConfigurationError(f"Site {name!r} has no 'type'")
        try:
            conn = site["connection"]
        except KeyError:
            raise ConfigurationError(f"Site {name!r} has no 'connection'")
        yield name, tp, conn



"""Base site class"""

import logging
import os
import pathlib

from .. import ConfigurationError, generator
from ..connection import PIPE
from ..utils import check_retcode, command_as_list, normalise_signal

_logger = logging.getLogger(__name__)


class Site:
    """Base site class

    Parameters
    ----------
    config: dict
        Site configuration

    connection: :py:class:`troika.connections.base.Connection`
        Connection object to interact with the site

    global_config: :py:class:`troika.config.Config`
        Global configuration
    """

    #: Value for the 'type' key in the site configuration.
    #: If None, the name will be computed by turning the class name to
    #: lowercase and removing a trailing "site" if present, e.g. ``FooSite``
    #: becomes ``foo``.
    __type_name__ = None

    #: Prefix for the generated directives, e.g. ``b"#SBATCH "``. If ``None``,
    #: no directives will be generated
    directive_prefix = None

    #: Directive translation table (``str`` -> ``bytes``). Values are formatted
    #: using the ``%`` operator
    directive_translate = {}

    def __init__(self, config, connection, global_config):
        self.config = config
        self._connection = connection
        try:
            self._kill_sequence = [
                (wait, normalise_signal(sig))
                for wait, sig in config.get("kill_sequence", [])
            ]
        except (TypeError, ValueError) as e:
            raise ConfigurationError(f"Invalid kill sequence: {e!s}")

    def submit(self, script, user, output, dryrun=False):
        """Submit a job

        The script and output path are interpreted according to the site.

        Parameters
        ----------
        script: path-like
            Path to the job script
        output: path-like
            Path to the output file
        user: str
            Remote user name
        dryrun: bool
            If True, do not submit, only report what would be done
        """
        raise NotImplementedError()

    def monitor(self, script, user, output=None, jid=None, dryrun=False):
        """Kill a submitted job

        The script and job ID are interpreted according to the site.
        If no job ID is provided, it will be inferred.

        Parameters
        ----------
        script: path-like
            Path to the job script
        user: str
            Remote user name
        output: path-like or None
            Path to the output file
        jid: str or None
            Job ID
        dryrun: bool
            If True, do not do anything, only report what would be done
        """
        raise NotImplementedError()

    def kill(self, script, user, output=None, jid=None, dryrun=False):
        """Kill a submitted or running job

        The script and job ID are interpreted according to the site.
        If no job ID is provided, it will be inferred.

        Parameters
        ----------
        script: path-like
            Path to the job script
        user: str
            Remote user name
        output: path-like or None
            Path to the output file
        jid: str or None
            Job ID
        dryrun: bool
            If True, do not kill, only report what would be done

        Returns
        -------
        tuple:
            [0]
                The job ID of the killed job
            [1]
                CANCELLED:
                    the job was cancelled before it started
                KILLED:
                    the job was killed while running without a
                    catchable signal allowing it to clean up or
                    report its demise
                TERMINATED:
                    the job was sent a catchable signal while running
                    and is expected to clean up and report its own
                    demise if necessary
                VANISHED:
                    the job has disappeared so no further attempt
                    could be made to kill it
        """
        raise NotImplementedError()

    def check_connection(self, timeout=None, dryrun=False):
        """Check whether the connection is working

        Parameters
        ----------
        timeout: int
            If set, consider the connection is not working if no response after
            this number of seconds
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed

        Returns
        -------
        bool
            True if the connection is able to execute commands
        """
        return self._connection.checkstatus(timeout=timeout, dryrun=dryrun)

    def create_output_dir(self, output, dryrun=False):
        """Create the output directory for a job to be submitted

        Parameters
        ----------
        output: path-like
            Path to the output file
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed

        Returns
        -------
        :py:class:`pathlib.PurePath`
            Path to the newly created directory
        """
        out_dir = pathlib.PurePath(output).parent
        pmkdir_command = command_as_list(
            self.config.get("pmkdir_command", ["mkdir", "-p"])
        )
        proc = self._connection.execute(
            pmkdir_command + [out_dir], stdout=PIPE, stderr=PIPE, dryrun=dryrun
        )
        if dryrun:
            return out_dir
        proc_stdout, proc_stderr = proc.communicate()
        if proc.returncode != 0:
            if proc_stdout:
                _logger.error("%s stdout:\n%s", pmkdir_command[0], proc_stdout.strip())
            if proc_stderr:
                _logger.error("%s stderr:\n%s", pmkdir_command[0], proc_stderr.strip())
            check_retcode(proc.returncode, what="Output directory creation")
        else:
            if proc_stdout:
                _logger.debug("%s stdout:\n%s", pmkdir_command[0], proc_stdout.strip())
            if proc_stderr:
                _logger.debug("%s stderr:\n%s", pmkdir_command[0], proc_stderr.strip())
        return out_dir

    def get_native_parser(self):
        """Create a :py:class:`troika.parser.Parser` for native directives

        Returns
        -------
        :py:class:`troika.parser.Parser` or None
            Directive parser, if any
        """
        return None

    def get_directive_translation(self):
        """Construct the translation params

        Returns
        -------
        tuple
            ``(directive_prefix, directive_translate)``, updated with the
            configuration overrides
        """
        prefix = self.config.get("directive_prefix", self.directive_prefix)
        translate = self.directive_translate.copy()
        for name, fmt in self.config.get("directive_translate", {}).items():
            if fmt is None:
                translate[name] = generator.ignore
            else:
                translate[name] = fmt.encode("utf-8")
        return (prefix, translate)

    def remove_previous_output(self, output, dryrun=False):
        """Remove previous output file if existing.

        Parameters
        ----------
        output: path-like
            Path to the output file
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed

        """
        if os.path.exists(output):
            if dryrun:
                _logger.info("removing:\n%s", output)
            else:
                os.remove(output)



"""Configuration file handling"""

import logging
import os
from collections import UserDict

import yaml

from . import ConfigurationError, InvocationError
from .utils import first_not_none

_logger = logging.getLogger(__name__)


class Config(UserDict):
    """Configuration mapping"""

    def get_site_config(self, name):
        """Get the configuration associated with the given site

        Parameters
        ----------
        name: str
            Name of the requested site

        Raises
        ------
        `ConfigurationError`
            if the configuration does not define a 'sites' object
        `KeyError`
            if the requested site is not defined
        """
        try:
            sites = self.data["sites"]
        except KeyError:
            raise ConfigurationError("No 'sites' defined in configuration")

        return sites[name]


def get_config(configfile=None, guesses=[]):
    """Read a configuration file

    If configfile is None, the path is read from the ``TROIKA_CONFIG_FILE``
    environment variable, or optional guesses.

    Parameters
    ----------
    configfile: None, path-like or file-like
        Configuration file (path or stream)
    guesses: list of path-like
        Try these paths as a last resort

    Returns
    -------
    `Config`
    """

    configfile = first_not_none(
        [
            configfile,
            os.environ.get("TROIKA_CONFIG_FILE"),
        ]
        + [guess for guess in guesses if os.path.exists(guess)]
    )
    if configfile is None:
        raise InvocationError("No configuration file found")

    try:
        path = os.fspath(configfile)
    except TypeError:  # not path-like
        pass
    else:
        configfile = open(path, "r")

    config_fname = configfile.name if hasattr(configfile, "name") else repr(configfile)
    _logger.debug("Using configuration file %s", config_fname)

    try:
        return Config(yaml.safe_load(configfile))
    except yaml.YAMLError as e:
        raise ConfigurationError(str(e))



"""Component discovery utilities"""

from importlib.metadata import entry_points


def get_entrypoint(group, name):
    """Load a component from a declared entry point

    Parameters
    ----------
    group: str
        Name of the entry point group, e.g. "troika.controllers"
    name: str
        Name of the component

    Returns
    -------
    module or object
        Component

    Raises
    ------
    KeyError
        If `group` is not found
    ValueError
        If `name` is not found in `group`
    """
    eps = entry_points()
    if hasattr(eps, "select"):  # Python >=3.10
        if group not in eps.groups:
            raise KeyError(group)
        found = eps.select(group=group, name=name)
    else:
        found = [comp for comp in eps[group] if comp.name == name]

    component = next(iter(found), None)
    if component is None:
        raise ValueError(f"Component {name!r} not found in group {group!r}")
    return component.load()



"""Slurm-managed site"""

import logging
import pathlib
import re
import signal
import time
from collections import OrderedDict

from .. import InvocationError, RunError, generator
from ..connection import PIPE
from ..parser import BaseParser, ParseError
from ..utils import check_retcode, command_as_list, parse_bool
from .base import Site

_logger = logging.getLogger(__name__)


def _split_slurm_directive(arg):
    """Split the argument of a Slurm directive

    >>> _split_slurm_directive(b"--output=foo")
    (b'--output', b'foo')
    >>> _split_slurm_directive(b"-J job")
    (b'-J', b'job')
    >>> _split_slurm_directive(b"--exclusive")
    (b'--exclusive', None)
    """
    m = re.match(rb"([^\s=]+)(=|\s+)?(.*)?$", arg)
    if m is None:
        raise ParseError(r"Malformed sbatch argument: {arg!r}")
    key, sep, val = m.groups()
    if sep is None:
        assert val == b""
        val = None
    return key, val


class SlurmDirectiveParser(BaseParser):
    """Parser that processes a script to extract Slurm directives

    Parameters
    ----------
    drop_keys: Iterable[bytes]
        Directives to ignore, e.g. ``[b'-o', b'--output']``

    Members
    -------
    data: collections.OrderedDict[bytes, (bytes or None, bytes)]
        Directives that have been parsed. The first item of the dict value is
        the parsed value, if any, and the second is the full line, including
        the line terminator.
    """

    DIRECTIVE_RE = re.compile(rb"^#\s*SBATCH\s+(.+)$")

    def __init__(self, drop_keys=None):
        super().__init__()
        self.data = OrderedDict()
        if drop_keys is None:
            drop_keys = []
        self.drop_keys = set(drop_keys)

    def feed(self, line):
        """Process the given line

        See ``BaseParser.feed``
        """
        m = self.DIRECTIVE_RE.match(line)
        if m is None:
            return False

        key, value = _split_slurm_directive(m.group(1))
        if key not in self.drop_keys:
            self.data[key] = (value, line)

        return True


def _translate_exclusive(value):
    if value == () or value == b"":
        value = True
    else:
        value = parse_bool(value, default=value)
    if not value:
        return None
    elif isinstance(value, bool):  # value is True
        return b"--exclusive"
    else:
        return b"--exclusive=%s" % value


def _translate_export_vars(value):
    if value in (b"all", b"none"):
        value = value.upper()
    return b"--export=%s" % value


def _translate_hyperthreading(value):
    if value == ():
        value = True
    else:
        value = parse_bool(value)
    flag = b"" if value else b"no"
    return b"--hint=%smultithread" % flag


def _translate_mail_type(value):
    trans = {b"none": b"NONE", b"begin": b"BEGIN", b"end": b"END", b"fail": b"FAIL"}
    vals = value.split(b",")
    newvals = []
    for val in vals:
        newval = trans.get(val.lower())
        if newval is None:
            _logger.warn("Unknown mail_type value %r", val)
            newval = val
        newvals.append(newval)
    return b"--mail-type=%s" % b",".join(newvals)


class SlurmSite(Site):
    """Site managed using Slurm"""

    directive_prefix = b"#SBATCH "
    directive_translate = {
        "billing_account": b"--account=%s",
        "cpus_per_task": b"--cpus-per-task=%s",
        "distribution": b"--distribution=%s",
        "enable_hyperthreading": _translate_hyperthreading,
        "error_file": b"--error=%s",
        "exclusive": _translate_exclusive,
        "export_vars": _translate_export_vars,
        "gpus_per_node": b"--gpus-per-node=%s",
        "gpus_per_task": b"--gpus-per-task=%s",
        "join_output_error": generator.ignore,
        "licenses": b"--licenses=%s",
        "mail_type": _translate_mail_type,
        "mail_user": b"--mail-user=%s",
        "memory_per_node": b"--mem=%s",
        "memory_per_cpu": b"--mem-per-cpu=%s",
        "name": b"--job-name=%s",
        "output_file": b"--output=%s",
        "partition": b"--partition=%s",
        "priority": b"--priority=%s",
        "reservation": b"--reservation=%s",
        "tasks_per_node": b"--ntasks-per-node=%s",
        "threads_per_core": b"--threads-per-core=%s",
        "tmpdir_size": b"--tmp=%s",
        "total_gpus": b"--gpus=%s",
        "total_nodes": b"--nodes=%s",
        "total_tasks": b"--ntasks=%s",
        "queue": b"--qos=%s",
        "walltime": b"--time=%s",
        "working_dir": b"--chdir=%s",
    }

    SUBMIT_RE = re.compile(rb"^(?:Submitted batch job )?(\d+)$", re.MULTILINE)

    def __init__(self, config, connection, global_config):
        super().__init__(config, connection, global_config)
        self._sbatch = command_as_list(config.get("sbatch_command", "sbatch"))
        self._scancel = command_as_list(config.get("scancel_command", "scancel"))
        self._squeue = command_as_list(config.get("squeue_command", "squeue"))
        self._copy_script = config.get("copy_script", False)
        self._copy_jid = config.get("copy_jid", False)

    def _parse_submit_output(self, out):
        match = self.SUBMIT_RE.search(out)
        if match is None:
            _logger.warn("Could not parse SLURM output %r", out)
            return None
        return int(match.group(1))

    def _get_state(self, jid, strict=True, dryrun=False):
        """Return the state of a SLURM job.

        Parameters
        ----------
        jid: int or str
            Job ID
        strict: bool
            If True (default), raise an exception for all failures of the
            squeue command. If False, ignore failures which are
            merely because the specified job does not exist, but raise an
            exception on all other failures.
        dryrun: bool
            If True (default False), do not do anything and return "DRYRUN"

        Returns
        -------
        str or None:
            Slurm job state, e.g. PENDING or RUNNING, if the job exists.
            Returns None if strict is in effect and squeue either outputs
            an empty string (as it can do for a job which has very recently
            disappeared), or fails with a message indicating that the job
            does not exist. If `dryrun` is True, the result is "DRYRUN"."""
        cmd = self._squeue + ["-h", "-o", "%T", "-j", str(jid)]
        proc = self._connection.execute(cmd, stdout=PIPE, stderr=PIPE, dryrun=dryrun)
        if dryrun:
            return "DRYRUN"
        proc_stdout, proc_stderr = proc.communicate()
        retcode = proc.returncode
        # Essential to remove trailing newline from stdout before returning
        proc_stdout = proc_stdout.strip()
        proc_stderr = proc_stderr.strip()
        _logger.debug("squeue output for job %d: %s", jid, proc_stdout)
        if retcode != 0:
            _logger.error("squeue error: %s", proc_stderr)
            # An intermediary (e.g. ecsbatch) may shift the error message to stdout rather than stderr
            if strict or all(
                b"Invalid job id specified" not in x for x in (proc_stdout, proc_stderr)
            ):
                check_retcode(retcode, what="Get State")
            else:
                return None
        else:
            if proc_stderr:
                _logger.debug("squeue error output: %s", proc_stderr)
            if strict and not proc_stdout:
                raise RunError(f"Get State for job {jid} produced no output")
        if proc_stdout:
            return proc_stdout.decode("ascii")

    def submit(self, script, user, output, dryrun=False):
        """See `troika.sites.Site.submit`"""
        script = pathlib.Path(script)

        cmd = self._sbatch.copy()

        if not script.exists():
            raise InvocationError(f"Script file {str(script)!r} does not exist")
        inpf = None
        if self._copy_script:
            script_remote = pathlib.PurePath(output).parent / script.name
            self._connection.sendfile(script, script_remote, dryrun=dryrun)
            cmd.append(script_remote)
        else:
            inpf = script.open(mode="rb")

        proc = self._connection.execute(
            cmd, stdin=inpf, stdout=PIPE, stderr=PIPE, dryrun=dryrun
        )
        if dryrun:
            return

        proc_stdout, proc_stderr = proc.communicate()
        if proc.returncode != 0:
            if proc_stdout:
                _logger.error(
                    "sbatch stdout for script %s:\n%s", script, proc_stdout.strip()
                )
            if proc_stderr:
                _logger.error(
                    "sbatch stderr for script %s:\n%s", script, proc_stderr.strip()
                )
            check_retcode(proc.returncode, what="submission")
        else:
            if proc_stdout:
                _logger.debug(
                    "sbatch stdout for script %s:\n%s", script, proc_stdout.strip()
                )
            if proc_stderr:
                _logger.debug(
                    "sbatch stderr for script %s:\n%s", script, proc_stderr.strip()
                )

        jobid = self._parse_submit_output(proc_stdout)
        _logger.debug("Slurm job ID: %d", jobid)

        jid_output = script.with_suffix(script.suffix + ".jid")
        if jid_output.exists():
            _logger.warning(
                "Job ID output file %r already exists, " + "overwriting",
                str(jid_output),
            )
        jid_output.write_text(str(jobid) + "\n")

        if self._copy_jid:
            jid_remote = pathlib.PurePath(output).parent / jid_output.name
            _logger.debug("Copying JID to output directory: %s", jid_remote)
            self._connection.sendfile(jid_output, jid_remote, dryrun=dryrun)

        return jobid

    def monitor(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.Site.monitor`"""
        script = pathlib.Path(script)

        if user is None:
            user = "$USER"

        if jid is None:
            jid = self._parse_jidfile(script, output)
            _logger.debug(f"Read job id {jid!r} from jidfile")
        else:
            _logger.debug(f"Using specified job id {jid!r}")
        try:
            jid = int(jid)
        except ValueError:
            raise RunError(f"Invalid job id: {jid!r}")

        stat_output = script.with_suffix(script.suffix + ".stat")
        if stat_output.exists():
            _logger.warning(
                "Status file %r already exists, overwriting", str(stat_output)
            )
        outf = None
        if not dryrun:
            outf = stat_output.open(mode="wb")

        self._connection.execute(
            self._squeue + ["-u", user, "-j", str(jid)], stdout=outf, dryrun=dryrun
        )

        _logger.info("Output written to %r", str(stat_output))

    def kill(self, script, user, output=None, jid=None, dryrun=False):
        """See `troika.sites.Site.kill`"""
        script = pathlib.Path(script)

        if jid is None:
            jid = self._parse_jidfile(script, output)
            _logger.debug(f"Read job id {jid!r} from jidfile")
        else:
            _logger.debug(f"Using specified job id {jid!r}")
        try:
            jid = int(jid)
        except ValueError:
            raise RunError(f"Invalid job id: {jid!r}")

        # Attempting to send a signal to a PENDING job will wait for it
        # to run first, which is not what we want. Cancel such jobs
        # directly, regardless of `_kill_sequence`. The "-t PENDING"
        # is important to prevent a race condition if the job is just
        # about to run.
        state = self._get_state(jid, strict=False, dryrun=dryrun)
        if state is None:
            # Job disappeared already
            return (jid, "VANISHED")
        elif state == "PENDING":
            cmd = self._scancel + ["-t", "PENDING", str(jid)]
            proc = self._connection.execute(cmd, stdout=PIPE, dryrun=dryrun)
            if not dryrun:
                proc_stdout, _ = proc.communicate()
                # Strip this _before_ checking for output because ecscancel
                # produces spurious blank lines
                proc_stdout = proc_stdout.strip()
                retcode = proc.returncode
                if retcode != 0:
                    _logger.error("scancel output: %s", proc_stdout)
                    if b"Invalid job id specified" in proc_stdout:
                        # Job disappeared already
                        return (jid, "VANISHED")
                    else:
                        check_retcode(retcode, what="Kill")
                elif proc_stdout:
                    _logger.debug("scancel output: %s", proc_stdout)

            state = self._get_state(jid, strict=False, dryrun=dryrun)
            if state is None or state == "CANCELLED":
                return (jid, "CANCELLED")
            elif state == "PENDING":
                if not dryrun:
                    raise RunError(f"Failed to cancel PENDING job {jid!r}")
            # If anything else, the job is probably starting, so fall through
            # and treat like a running job

        seq = self._kill_sequence
        if not seq:
            seq = [(0, None)]

        cancel_status = None
        for wait, sig in seq:
            time.sleep(wait)

            cmd = self._scancel + [str(jid)]
            if sig is not None:
                cmd.extend(["-f", "-s", str(sig.value)])
            proc = self._connection.execute(cmd, stdout=PIPE, dryrun=dryrun)

            if dryrun:
                continue

            proc_stdout, _ = proc.communicate()
            retcode = proc.returncode
            # Strip this _before_ checking for output because ecscancel
            # produces spurious blank lines
            proc_stdout = proc_stdout.strip()
            if retcode != 0:
                if cancel_status is None:
                    _logger.error("scancel output: %s", proc_stdout)
                    if b"Invalid job id specified" in proc_stdout:
                        # Job disappeared already
                        return (jid, "VANISHED")
                    else:
                        check_retcode(retcode, what="Kill")
                else:
                    _logger.debug("scancel output: %s", proc_stdout)
                    break
            elif proc_stdout:
                _logger.debug("scancel output: %s", proc_stdout)

            if sig is None or sig == signal.SIGKILL:
                cancel_status = "KILLED"
            elif cancel_status is None:
                cancel_status = "TERMINATED"
        return (jid, cancel_status)

    def get_native_parser(self):
        """See `troika.sites.Site.get_native_parser`"""
        return SlurmDirectiveParser(drop_keys=[b"-o", b"--output", b"-e", b"--error"])

    def _parse_jidfile(self, script, output=None, dryrun=False):
        script = pathlib.Path(script)
        jid_output = script.with_suffix(script.suffix + ".jid")
        try:
            return jid_output.read_text().strip()
        except IOError as e:
            if self._copy_jid and output is not None:
                jid_remote = pathlib.PurePath(output).parent / jid_output.name
                try:
                    self._connection.getfile(jid_remote, jid_output, dryrun=dryrun)
                    _logger.debug(
                        "Job ID file copied back from output directory: %s", jid_remote
                    )
                    if not dryrun:
                        return jid_output.read_text().strip()
                except (IOError, RunError) as e2:
                    raise RunError(
                        f"Could not read the job id: {e!s} or copy it back {e2!s}"
                    )
            raise RunError(f"Could not read the job id: {e!s}")

    def __repr__(self):
        return f"{self.__class__.__name__}(connection={self._connection!r}, sbatch_command={self._sbatch[0]!r})"



"""Base controller class"""

import logging
import os
import pathlib
import shutil
import tempfile

from .. import ConfigurationError, InvocationError, RunError, hook, site
from ..directives import ALIASES, translators
from ..generator import Generator
from ..parser import DirectiveParser, MultiParser, ParseError, ShebangParser

_logger = logging.getLogger(__name__)


class Controller:
    """Main controller

    Parameters
    ----------
    config: :py:class:`troika.config.Config`
        Configuration
    args: :py:class:`argparse.Namespace`
        Command-line arguments
    logfile: path-like or None
        Path to the log file
    """

    __type_name__ = "base"

    def __init__(self, config, args, logfile):
        self.config = config
        self.args = args
        self.logfile = logfile
        self.site = None
        self.default_shebang = None
        self.unknown_directive = "warn"
        self.script_data = {}

    def __repr__(self):
        """Return a printable representation"""
        return "Controller"

    def submit(self, script, user, output, dryrun=False):
        """Process a 'submit' command

        The script and job ID are interpreted according to the site.

        Parameters
        ----------
        script: path-like
            Path to the job script
        user:
            Remote user name
        output: path-like
            Path to the job output file
        dryrun: bool
            If True, do not submit, only report what would be done

        Returns
        -------
        int
            Return code (0 for success)
        """
        with self.action_context(parse_script=script) as context:
            pp_script = self.generate_script(script, user, output)
            hook.pre_submit(self.site, script, output, dryrun)
            self.site.submit(pp_script, user, output, dryrun)
        return context.status

    def monitor(self, script, user, output=None, jid=None, dryrun=False):
        """Process a 'monitor' command

        The script and job ID are interpreted according to the site.
        If no job ID is provided, it will be inferred.

        Parameters
        ----------
        script: path-like
            Path to the job script
        user: str
            Remote user name
        output: path-like or None
            Path to the job output file
        jid: str or None
            Job ID
        dryrun: bool
            If True, do not do anything, only report what would be done

        Returns
        -------
        int
            Return code (0 for success)
        """
        with self.action_context() as context:
            self.site.monitor(script, user, output, jid, dryrun)
        return context.status

    def kill(self, script, user, output=None, jid=None, dryrun=False):
        """Process a 'kill' command

        The script and job ID are interpreted according to the site.
        If no job ID is provided, it will be inferred.

        Parameters
        ----------
        script: path-like
            Path to the job script
        user: str
            Remote user name
        output: path-like or None
            Path to the job output file
        jid: str or None
            Job ID
        dryrun: bool
            If True, do not kill, only report what would be done

        Returns
        -------
        int
            Return code (0 for success)
        """
        with self.action_context() as context:
            jid, cancel_status = self.site.kill(script, user, output, jid, dryrun)
            hook.post_kill(self.site, script, output, jid, cancel_status, dryrun)
        return context.status

    def check_connection(self, timeout=None, dryrun=False):
        """Process a 'check-connection' command

        Parameters
        ----------
        timeout: int
            If set, consider the connection is not working if no response after
            this number of seconds
        dryrun: bool
            If True, do not do anything but print the command that would be
            executed

        Returns
        -------
        bool
            True if the connection is able to execute commands
        """
        with self.action_context() as context:
            working = self.site.check_connection(timeout=timeout, dryrun=dryrun)
            if not working:
                context.status = 1
        return context.status == 0

    def list_sites(self):
        """Process a 'list-sites' command

        Yields
        -------
        ``(name, type, connection)``
        """
        yield from site.list_sites(self.config)

    class ActionContext:
        def __init__(self, controller, *args, **kwargs):
            self._controller = controller
            self._controller.setup(*args, **kwargs)

        def __enter__(self):
            self.status = None
            return self

        def __exit__(self, exc_type, exc, traceback):
            if exc is None:
                if self.status is None:
                    self.status = 0
                swallow = False
            else:
                if self.status is None or self.status == 0:
                    self.status = 1
                if isinstance(exc, ConfigurationError):
                    _logger.critical("Configuration error: %s", exc)
                elif isinstance(exc, InvocationError):
                    _logger.critical("Invocation error: %s", exc)
                elif isinstance(exc, RunError):
                    _logger.critical("%s", exc)
                else:
                    _logger.error(
                        "Unhandled exception", exc_info=(exc_type, exc, traceback)
                    )
                swallow = True
            try:
                self._controller.teardown(self.status)
            except Exception as e:
                # An error during the at-exit hooks should _not_ be reported as
                # failure of the operation (e.g. submission failure)
                _logger.error("Exception during teardown/exit", exc_info=e)
            return swallow

    def action_context(self, *args, **kwargs):
        """Create a context manager for executing an action

        The arguments are passed to :py:meth:`setup` when entering the context manager
        """
        return self.ActionContext(self, *args, **kwargs)

    def setup(self, parse_script=None):
        """Set up the controller

        Parameters
        ----------
        parse_script: path-like, optional
            Parse this script
        """
        self.site = self._get_site()
        if parse_script is not None:
            self.parse_script(parse_script)
        hook.setup_hooks(self.config, self.args.site)
        res = hook.at_startup(self.args.action, self.site, self.args)
        if any(res):
            raise SystemExit(1)
        self.default_shebang = self.site.config.get("default_shebang", None)
        self.unknown_directive = self.site.config.get("unknown_directive", "warn")

    def teardown(self, sts=0):
        """Tear down the controller

        Parameters
        ----------
        sts: int
            Exit status
        """
        hook.at_exit(self.args.action, self.site, self.args, sts, self.logfile)

    def parse_script(self, script):
        """Parse the script

        Parameters
        ----------
        script: path-like
            Script to parse
        """
        dir_parser = DirectiveParser(aliases=ALIASES)
        parsers = [("directives", dir_parser)]
        native = self.site.get_native_parser()
        if native is not None:
            parsers.append(("native", native))
        parsers.append(("shebang", ShebangParser()))
        parser = MultiParser(parsers)
        body = self.run_parser(script, parser)
        self.script_data.update(parser.data)
        self.script_data["body"] = body
        dir_defines = getattr(self.args, "define", [])
        dir_overrides = dir_parser.parse_directive_args(dir_defines)
        self.script_data["directives"].update(dir_overrides)

    def run_parser(self, script, parser):
        """Run the parser on the script

        Parameters
        ----------
        script: path-like
            Path to the script
        parser: :py:class:`troika.parser.Parser`
            Parser to use

        Returns
        -------
        file-like
            Script body
        """
        script = pathlib.Path(script)
        stmp = tempfile.SpooledTemporaryFile(
            max_size=1024**3, mode="w+b", dir=script.parent, prefix=script.name
        )
        with open(script, "rb") as sin:
            try:
                for lineno, line in enumerate(sin, start=1):
                    drop = parser.feed(line)
                    if not drop:
                        stmp.write(line)
            except ParseError as e:
                raise ParseError(f"in {script!s}, line {lineno} {e!s}") from e
        stmp.seek(0)
        return stmp

    def generate_script(self, script, user, output):
        """Generate the post-processed script

        Parameters
        ----------
        script: path-like
            Path to the script
        user: str or None
            Remote user name
        output: path-like
            Path to the job output file

        Returns
        -------
        path-like
            Path to the post-processed script
        """
        if (
            self.default_shebang is not None
            and self.script_data.get("shebang", None) is None
        ):
            self.script_data["shebang"] = self.default_shebang.encode("utf-8")
        directive_prefix, directive_translate = self.site.get_directive_translation()
        generator = Generator(
            directive_prefix, directive_translate, self.unknown_directive
        )
        self.script_data["directives"]["output_file"] = os.fsencode(output)
        self.script_data = translators(self.script_data, self.config, self.site)
        return self.run_generator(script, generator)

    def run_generator(self, script, generator):
        """Generate the script file

        Parameters
        ----------
        script: path-like
            Path to the script
        generator: :py:class:`troika.generator.Generator`
            Generator to run

        Returns
        -------
        path-like
            Path to the generated script file
        """
        script = pathlib.Path(script)
        orig_script = script.with_suffix(script.suffix + ".orig")
        if orig_script.exists():
            _logger.warning(
                "Backup script file %r already exists, " + "overwriting",
                str(orig_script),
            )
        with tempfile.NamedTemporaryFile(
            mode="w+b", delete=False, dir=script.parent, prefix=script.name
        ) as sout:
            sout.writelines(generator.generate(self.script_data))
            for line in self.script_data["body"]:
                sout.write(line)
            new_script = pathlib.Path(sout.name)
        shutil.copymode(script, new_script)
        shutil.copy2(script, orig_script)
        new_script.replace(script)
        _logger.debug("Script generated. Original script saved to %r", str(orig_script))
        return script

    def _get_site(self):
        """Select the site and create the corresponding :py:class:`troika.sites.base.Site` instance"""
        return site.get_site(self.config, self.args.site, self.args.user)



"""Common hooks"""

import logging
import pathlib

_logger = logging.getLogger(__name__)


def ensure_output_dir(site, output, dryrun=False):
    """Ensure the output directory exists and return its path"""
    return site.create_output_dir(output, dryrun=dryrun)


def remove_previous_output(site, script, output, dryrun=False):
    """Pre-submit hook to potentially previous output"""
    return site.remove_previous_output(output, dryrun=dryrun)


def check_connection(action, site, args):
    """Startup hook to check the connection works before doing anything"""
    working = site.check_connection(dryrun=args.dryrun)
    if not working:
        _logger.error("Connection not working. Exiting.")
        return True


def create_output_dir(site, script, output, dryrun=False):
    """Pre-submit hook to create the output directory"""
    ensure_output_dir(site, output, dryrun=dryrun)


def copy_orig_script(site, script, output, dryrun=False):
    """Pre-submit hook to copy the original script to the remote server"""
    out_dir = ensure_output_dir(site, output, dryrun=dryrun)
    script = pathlib.PurePath(script)
    orig_script = script.with_suffix(script.suffix + ".orig")
    site._connection.sendfile(orig_script, out_dir / orig_script.name, dryrun=dryrun)


def copy_submit_logfile(action, site, args, sts, logfile):
    """Exit hook to copy the log file to the remote server when submitting a job"""
    if action != "submit":
        return
    logfile = pathlib.PurePath(logfile)
    out_dir = ensure_output_dir(site, args.output, dryrun=args.dryrun)
    site._connection.sendfile(logfile, out_dir / logfile.name, dryrun=args.dryrun)


def copy_kill_logfile(action, site, args, sts, logfile):
    """Exit hook to copy the log file to the remote server when killing a job"""
    if action != "kill":
        return
    if args.output:
        logfile = pathlib.PurePath(logfile)
        out_dir = ensure_output_dir(site, args.output, dryrun=args.dryrun)
        site._connection.sendfile(logfile, out_dir / logfile.name, dryrun=args.dryrun)
    else:
        _logger.error("copy_kill_logfile hook requires output argument to be passed")



"""Main controller"""

import logging

from . import ConfigurationError
from .components import get_entrypoint

_logger = logging.getLogger(__name__)


def get_controller(config, args, logfile):
    """Create a `troika.controllers.base.Controller` object from configuration"""

    ctl = config.get("controller", "base")
    try:
        cls = get_entrypoint("troika.controllers", ctl)
    except ValueError:
        raise ConfigurationError(f"Unknown controller {ctl!r}")

    ctl = cls(config, args, logfile)
    _logger.debug("Using %r", ctl)
    return ctl
